/**
    @file ft51_usb_transfer.c

    @brief USB Library API transfer functions.

    @version 1.0.0

**/
/*
 * ============================================================================
 * History
 * =======
 *
 * (C) Copyright 2014, Future Technology Devices International Ltd.
 * ============================================================================
 *
 * This source code ("the Software") is provided by Future Technology Devices
 * International Limited ("FTDI") subject to the licence terms set out
 * http://www.ftdichip.com/FTSourceCodeLicenceTerms.htm ("the Licence Terms").
 * You must read the Licence Terms before downloading or using the Software.
 * By installing or using the Software you agree to the Licence Terms. If you
 * do not agree to the Licence Terms then do not download or use the Software.
 *
 * Without prejudice to the Licence Terms, here is a summary of some of the key
 * terms of the Licence Terms (and in the event of any conflict between this
 * summary and the Licence Terms then the text of the Licence Terms will
 * prevail).
 *
 * The Software is provided "as is".
 * There are no warranties (or similar) in relation to the quality of the
 * Software. You use it at your own risk.
 * The Software should not be used in, or for, any medical device, system or
 * appliance. There are exclusions of FTDI liability for certain types of loss
 * such as: special loss or damage; incidental loss or damage; indirect or
 * consequential loss or damage; loss of income; loss of business; loss of
 * profits; loss of revenue; loss of contracts; business interruption; loss of
 * the use of money or anticipated savings; loss of information; loss of
 * opportunity; loss of goodwill or reputation; and/or loss of, damage to or
 * corruption of data.
 * There is a monetary cap on FTDI's liability.
 * The Software may have subsequently been amended by another user and then
 * distributed by that other user ("Adapted Software").  If so that user may
 * have additional licence terms that apply to those amendments. However, FTDI
 * has no liability in relation to those amendments.
 * ============================================================================
 */
#include <stdint.h>
#include <stdlib.h>
#include "ft51.h"
#include "ft51_sfr.h"
#include "ft51_usb.h"
#include "ft51_usb_internal.h"

uint8_t  USB_in_request(uint8_t ep_index, uint8_t *data, uint16_t length);
uint16_t USB_out_request(uint8_t ep_index, uint8_t *data, uint16_t length);

/**
 @brief   \par Detect DATA phase of a SETUP with OUT data
 @details Return a non-zero value when the DATA phase of a SETUP
 OUT packet is received.
 **/
static uint8_t USB_setup_out_data_recv(void)
{
    uint8_t process = 0;

    DISABLE_INTERRUPTS;
    if (USB_ep[CTRL_OUT].last_status)
    {
        process = 1;
        USB_ep[CTRL_OUT].last_status = 0;
    }
    ENABLE_INTERRUPTS;

    return process;
}

/**
 @brief   \par Detect DATA phase of a SETUP with IN data
 @details Return a non-zero value when the DATA phase of a SETUP
 IN packet is received.
 **/
static uint8_t USB_setup_in_data_recv(void)
{
    uint8_t process = 0;

    DISABLE_INTERRUPTS;
    if (USB_ep[CTRL_IN].last_status)
    {
        process = 1;
        USB_ep[CTRL_IN].last_status = 0;
    }
    ENABLE_INTERRUPTS;

    return process;
}

/**
 @brief \par USB IN Request
 @details Writes the data to the USB hardware.
 @param[in] ep Endpoint to send the IN request to.
 @param[in] data The command to write to FT122_CMD.
 @param[in] length Number of bytes to read.
 @return 1 for success, 0 for failure.
 **/
uint8_t USB_in_request(uint8_t ep_index, uint8_t *data, uint16_t length)
{
    uint16_t bytes_written = length;
    uint16_t attempts;
    uint8_t state = 0;

    // Before writing to the endpoint, wait for one of its buffers to become
    // empty.  SELECT_BUFFER_FULL indicates that both buffers are full (for a
    // non-Control endpoint), or, for a Control endpoint, that the single
    // buffer is full.
    for (attempts = 0; attempts < 100; attempts++)
    {
        DISABLE_INTERRUPTS;
        FT122_CMD = SELECT_ENDPOINT + ep_index;
        state = FT122_DATA;
        ENABLE_INTERRUPTS;
        if ((state & SELECT_BUFFER_FULL) == 0x00)
            break; // Buffer is empty, so stop checking.
    }

    if (state & SELECT_BUFFER_FULL)
    {
        return 0; // Buffer still full, so write nothing.
    }

    DISABLE_INTERRUPTS;

    FT122_CMD = ep_index; // Select Endpoint
    FT122_CMD = WR_BUFFER; // Write Buffer Command

    FT122_DATA = MSB(length);
    FT122_DATA = LSB(length);

    while (length--)
        FT122_DATA = *data++;

    // Validate buffer so that the data leaves the chip!
    FT122_CMD = VALIDATE_BUFFER;

    // Update ep state BUFFER_0_FULL etc. flags
    FT122_CMD = READ_ENDPOINT_STATUS + ep_index;
    USB_ep[ep_index].status = FT122_DATA;

    ENABLE_INTERRUPTS;

    return 1; // All bytes written
}

/**
 @brief \par USB OUT Request
 @details Reads data from the USB hardware.
 @param[in] ep Endpoint to read data from.
 @param[in] data Suitably-sized buffer to hold the data.
 @return Returns the number of bytes read.
 **/
uint16_t USB_out_request(uint8_t ep_index, uint8_t *data, uint16_t length)
{
    uint16_t buff_size = 0;
    uint16_t bytes_read = 0;
    uint8_t state = 0;

    DISABLE_INTERRUPTS;

    FT122_CMD = ep_index;
    state = FT122_DATA;

    if (state & USB_SELECT_ENDPOINT_BUFFER_FULL)
    {
        FT122_CMD = RD_BUFFER;

        // Length
        buff_size = (FT122_DATA << 8) & 0xff00; // MSB
        buff_size |= FT122_DATA & 0x00ff;       // LSB

        // Only read as many bytes as we have space for.
        if (buff_size > length)
            buff_size = length;

        bytes_read = buff_size;

        while (buff_size--)
            *data++ = FT122_DATA;

        FT122_CMD = CLR_BUFFER;

        // Update ep status BUFFER_0_FULL etc. flags
        FT122_CMD = READ_ENDPOINT_STATUS + ep_index;
        USB_ep[ep_index].status = FT122_DATA;
    }

    ENABLE_INTERRUPTS;

    return bytes_read;
}

/**
 @brief   \par Perform an IN or OUT data transfer on an endpoint
 @details Return number of bytes transferred.
 **/
uint16_t USB_transfer(USB_ENDPOINT_NUMBER ep_number, USB_ENDPOINT_DIR ep_dir,
        uint8_t *data, const uint16_t length)
{
    uint8_t ep_index = (ep_number * 2) + ep_dir;
    USB_endpoint *ep = &USB_ep[ep_index];
    uint16_t packet_len;
    uint16_t remaining_len = length;
    uint8_t *pdata = data;
    uint16_t max_bytes = USB_ep_size_bytes[ep->max_packet_size];
    uint16_t transferred = 0;
    uint8_t  sending_zlp = 0;

    for ( ; ; )
    {
        packet_len = remaining_len;
        if (packet_len > max_bytes)
        {
            packet_len = max_bytes;
        }

        if (ep->direction == USB_DIR_OUT)
        {
            if (ep->type == USB_EP_CTRL)
            {
                while (USB_setup_out_data_recv() == 0)
                {
                    // Interrupt OUT SETUP transfer due to IN packet from host
                    if (USB_setup_in_data_recv())
                    {
                        goto exit;
                    }
                }
            }

            packet_len = USB_out_request(ep_index, pdata, packet_len);
            transferred += packet_len;

            if (packet_len < max_bytes)
            {
                // Nothing to read, or short packet.
                // Either way, data-receiving is complete.
                break;
            }
        }
        else
        {
            if (ep->type == USB_EP_CTRL)
            {
                // Interrupt IN SETUP transfer due to OUT packet from host
                if (USB_setup_out_data_recv())
                {
                    goto exit;
                }
            }

            if (USB_in_request(ep_index, pdata, packet_len))
            {
                transferred += packet_len;
                if (sending_zlp)
                    break; // We've sent the extra ZLP so we're finished.
            }
            else
            {
                // Didn't write packet because Host hasn't read previous one.
                // Go round the loop and try again.
                continue;
            }
        }

        /* Move pointer to next chunk of data to send or overwrite. */
        pdata += packet_len;

        /* Decrement the total length remaining in the transfer. */
        remaining_len -= packet_len;

        if (remaining_len == 0)
        {
            // Transfer complete.
            if (ep->direction == USB_DIR_IN &&
                ep->type != USB_EP_INT &&
                packet_len != 0 && 
                packet_len % max_bytes == 0)
            {
                /* Non-Interrupt IN endpoints need to append a zero-length
                 * packet to tell the host that data-sending is complete, if
                 * the data length is an exact multiple of the max packet size.
                 */
                sending_zlp = 1;
            }
            else
            {
                break; // Nothing further to do.
            }
        }
    }

exit:    
    return transferred;
}
